"""
Tree cascades
=============

This example shows how to use `iplotx` to add cascading backgrounds to trees.
"Cascading" here means that each patch (rectangle/wedge/etc.) will cover a node
and all descendants, down to the leaves.
"""

from Bio import Phylo
from io import StringIO
import matplotlib.pyplot as plt
import iplotx as ipx

# Make a tree from a string in Newick format
tree = next(
    Phylo.NewickIO.parse(
        StringIO(
            "(()(()((()())(()()))))",
        )
    )
)

backgrounds = {
    tree.get_nonterminals()[3]: "turquoise",
    tree.get_terminals()[0]: "tomato",
    tree.get_terminals()[1]: "purple",
}

ipx.plotting.tree(
    tree,
    cascade_facecolor=backgrounds,
)

# %%
# Cascading patches have a style option "extend" which affects whether the patches extend to the
# end of the deepest leaf:

ipx.plotting.tree(
    tree,
    layout="vertical",
    cascade_facecolor=backgrounds,
    cascade_extend=True,
)

# %%
# Cascading patches work with radial layouts as well:

# sphinx_gallery_thumbnail_number = 3
ipx.plotting.tree(
    tree,
    layout="radial",
    cascade_facecolor=backgrounds,
    cascade_extend=True,
    aspect=1,
)

# %%
# Cascading patches exclude leaf labels by default, but can be extended to cover leaf labels by
# setting the parameter as follows:

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(9, 5))
ipx.plotting.tree(
    tree,
    ax=ax1,
    layout="radial",
    leaf_labels={leaf: f"L{i + 1}" for i, leaf in enumerate(tree.get_terminals())},
    cascade_facecolor=backgrounds,
    cascade_extend=True,  # Exclude leaf labels
    aspect=1,
    margins=0.4,
    title="Exclude leaf labels",
)
ipx.plotting.tree(
    tree,
    ax=ax2,
    layout="radial",
    leaf_labels={leaf: f"L{i + 1}" for i, leaf in enumerate(tree.get_terminals())},
    cascade_facecolor=backgrounds,
    cascade_extend="leaf_labels",  # Include leaf labels
    aspect=1,
    margins=0.4,
    title="Include leaf labels",
)

# %%
# This extension also works with other layouts, such as horizontal and vertical:

layout_and_orientations = {
    "horizontal": ["left", "right"],
    "vertical": ["descending", "ascending"],
    "radial": ["clockwise", "counterclockwise"],
}

fig, axs = plt.subplots(3, 2, figsize=(9, 13.5))
for i, (layout_name, layout_orientations) in enumerate(layout_and_orientations.items()):
    for j, orientation in enumerate(layout_orientations):
        ipx.plotting.tree(
            tree,
            ax=axs[i, j],
            layout=layout_name,
            layout_orientation=orientation,
            leaf_labels={leaf: f"L{i + 1}" for i, leaf in enumerate(tree.get_terminals())},
            cascade_facecolor=backgrounds,
            cascade_extend="leaf_labels",
            aspect=1,
            title=f"{layout_name} ({orientation})",
            margins=0.1,
        )



"""
Style hierarchy
===============

This example shows a few elements from  ``iplotx``'s style hierarchy.
"""

import matplotlib.pyplot as plt
import iplotx as ipx

tree = {
    "name": "style",
    "children": [
        {
            "name": "vertex",
            "children": [
                {
                    "name": "size",
                },
                {"name": "..."},
            ],
        },
        {"name": "..."},
    ],
}
tree = ipx.ingest.providers.tree.simple.SimpleTree.from_dict(tree)

fig, ax = plt.subplots(figsize=(4, 2.5))
ipx.tree(
    tree,
    ax=ax,
    vertex_labels=True,
    vertex_size=[(50, 25)] * 5,
    vertex_facecolor="none",
    vertex_edgecolor="black",
    vertex_linewidth=1.5,
    vertex_marker="r",
    vertex_label_horizontalalignment="center",
    vertex_label_hmargin=0,
    margins=(0.1, 0.1),
)
ax.invert_yaxis()



import importlib
import unittest
import pytest
import numpy as np
import pandas as pd
from iplotx.ingest import heuristics

if importlib.util.find_spec("igraph") is None:
    raise unittest.SkipTest("igraph not found, skipping tests")
else:
    import igraph as ig
if importlib.util.find_spec("networkx") is None:
    raise unittest.SkipTest("networkx not found, skipping tests")
else:
    import networkx as nx


def test_empty_layout():
    res = heuristics.normalise_layout(None, network={}, nvertices=0)
    assert isinstance(res, pd.DataFrame)
    assert res.shape == (0, 2)

    res = heuristics.normalise_layout(None, network=None, nvertices=0)
    assert res is None


def test_array_layout():
    res = heuristics.normalise_layout(np.zeros((3, 2)), network={}, nvertices=3)
    assert isinstance(res, pd.DataFrame)
    assert res.shape == (3, 2)


def test_igraph_layouts():
    g = ig.Graph.Ring(5)
    g["circle"] = g.layout_circle()
    res = heuristics.normalise_layout("circle", network=g)
    assert isinstance(res, pd.DataFrame)
    assert res.shape == (5, 2)

    g = ig.Graph.Ring(5)
    res = heuristics.normalise_layout("circle", network=g)
    assert isinstance(res, pd.DataFrame)
    assert res.shape == (5, 2)

    g = ig.Graph.Ring(5)
    layout = g.layout_circle()
    res = heuristics.normalise_layout(layout, network=g)
    assert isinstance(res, pd.DataFrame)
    assert res.shape == (5, 2)


def test_networkx_layouts():
    G = nx.house_graph()
    nx.set_node_attributes(G, {0: (0, 0), 1: (1, 0), 2: (0, 1), 3: (1, 1), 4: (0.5, 2.0)}, "pos")
    res = heuristics.normalise_layout("pos", network=G)
    assert isinstance(res, pd.DataFrame)
    assert res.shape == (5, 2)


accepted_groupings = [
    [{0, 1}, {0, 2}, {2, 4}],
    [0, 1, 0, 2, 2],
    {0: {0, 1}, 1: {0, 2}},
    {0: 0, 1: 1, 2: 0, 3: 0},
    ig.clustering.Clustering([0, 0, 1, 2, 2]),
    ig.clustering.Cover([[0, 1], [0, 1, 2], [2, 4]]),
]


@pytest.mark.parametrize("grouping", accepted_groupings)
def test_grouping_dict(grouping):
    g = ig.Graph.Ring(5)
    layout = g.layout_circle()
    res = heuristics.normalise_grouping(
        grouping,
        layout,
    )
    assert isinstance(res, dict)
    for k, v in res.items():
        assert isinstance(k, int)
        assert isinstance(v, (set, frozenset))


def test_grouping_sequence_invalid():
    g = ig.Graph.Ring(5)
    layout = g.layout_circle()
    grouping = [[0], 1, 0, 2, 2]
    with pytest.raises(TypeError):
        heuristics.normalise_grouping(
            grouping,
            layout,
        )



"""
Tree of trees
=============

This example shows how to plot trees at the leaf of a tree, using the `iplotx` library.
Note that there are a few conceivable ways to do similar things as this in
`matplotlib` and this is just one such approach, using inset axes.
"""

from ete4 import Tree
import matplotlib.pyplot as plt
import iplotx as ipx

tree = Tree(
    "(,((),(((),()),((),()))));",
)
subtree_strings = [
    "((a,b),(c,d));",
    "(a,((b,c),d));",
    "(a,b,c,d);",
    "(a,(b,(c,d)));",
    "(a,b);",
    "(a,(b,c,d));",
]

# Plot the initial tree
fig, ax = plt.subplots(figsize=(8, 5))
art = ipx.tree(
    tree,
    ax=ax,
    aspect=1,
    edge_color="grey",
)

for i, leaf in enumerate(tree.leaves()):
    # Create an axis in the rough proximity
    # and with the rough diemension of a leaf,
    # with some padding
    x, y = art.get_layout().T[leaf].values
    subax = ax.inset_axes(
        (x + 0.1, y - 0.4, 1.2, 0.8),
        transform=ax.transData,
    )

    # For each leaf, make a tree
    subtree = Tree(subtree_strings[i])
    # Plot the tree with dashed lines
    ipx.tree(
        subtree,
        ax=subax,
        edge_linestyle="--",
    )

# %%
# Combining this with other elements, e.g. cascading backgrounds, is possible too:

fig, ax = plt.subplots(figsize=(8, 5))
art = ipx.tree(
    tree,
    ax=ax,
    aspect=1,
    edge_color="grey",
    cascade_facecolor={
        list(tree.leaves())[0]: "lightblue",
        list(tree.traverse("preorder"))[2]: "pink",
    },
)
for i, leaf in enumerate(tree.leaves()):
    x, y = art.get_layout().T[leaf].values
    subax = ax.inset_axes(
        (x + 0.1, y - 0.4, 1.2, 0.8),
        transform=ax.transData,
    )
    # Set the inset axes background transparent
    subax.patch.set_alpha(0)

    subtree = Tree(subtree_strings[i])
    ipx.tree(
        subtree,
        ax=subax,
        edge_linestyle="--",
    )



"""
Edge shrink
===========

This example illustrates how to shrink edges, i.e. leave a bit of empty space between
edge cap (end) and the border of its source/target vertices.
"""

import iplotx as ipx

graph = {
    "edges": [
        ("A", "B"),
        ("B", "C"),
        ("C", "D"),
        ("C", "E"),
        ("E", "Bingo"),
    ],
    "directed": True,
}
layout = {
    "A": (0, 0),
    "B": (1, 0),
    "C": (1, 1),
    "D": (-1, 1),
    "E": (-0.5, 2),
    "Bingo": (0, 3),
}

ipx.network(
    graph,
    layout,
    style="rededge",
    edge_shrink=5,
    vertex_labels=True,
    vertex_size={"A": 20, "B": 20, "C": 20, "D": 20, "E": 20, "Bingo": 50},
)



"""
Zero-dependency trees
=====================

This example demonstrates how to visualise trees using `iplotx`'s internal `SimpleTree`,
a simple tree representation via the ``children`` attribute.

.. tip::
    This provider is not as powerful as proper tree libraries such as ``cogent3`` or ``ETE4``. It
    can be especially useful when you do not want to add dependencies or for educational purposes.
"""

import iplotx as ipx

tree = {
    "children": (
        {},
        {
            "children": (
                {
                    "children": (
                        {"branch_length": 1.5},
                        {}
                    )
                },
                {
                    "children": (
                        {},
                        {}
                    )
                }
            )
        }
    )
}

# Convert to our simple tree data structure
tree = ipx.ingest.providers.tree.simple.SimpleTree.from_dict(tree)

ipx.tree(
    tree,
)

# %%
# .. note::
#     ``iplotx`` does not generally provide network or tree data structures.
#     This is an exception to demonstrate how to visualise trees without
#     external dependencies.



"""
The red and the black
=====================

This example demonstrates how to overlay an ``iplotx`` artist with plain matplotlib patches.
"""


import numpy as np
import iplotx as ipx
import matplotlib.pyplot as plt

# We will use a network laid out onto a square grid
gsize = 5  # Grid size
layout = [(i % gsize, i // gsize) for i in range(gsize**2)]

# Two possible colors
colorlist = np.array(["tomato", "#111"])

# Assign node colors at random
node_colors = colorlist[np.random.randint(2, size=gsize**2)]

# Connect two neighboring nodes if they have the same color
edges = []
for inode in range(gsize**2):
    i = inode // gsize
    j = inode % gsize
    # Four possible edges
    cands = []
    if i != 0:
        cands.append(inode - gsize)
    if j != 0:
        cands.append(inode - 1)
    if i != gsize - 1:
        cands.append(inode + gsize)
    if j != gsize - 1:
        cands.append(inode + 1)
    for inode2 in cands:
        if node_colors[inode] == node_colors[inode2]:
            if (inode2, inode) not in edges:
                edges.append((inode, inode2))

# Build the network using our zero-dependency data structure
network = {
    "edges": edges,
    "nodes": list(range(gsize**2)),
}

# Visualise the network
fig, ax = plt.subplots(figsize=(5, 5))
ipx.network(
    network,
    layout=layout,
    ax=ax,
    vertex_size=8,
    vertex_marker='d',
)

# Add square patches centered on each node
for i in range(gsize):
    for j in range(gsize):
        inode = i * gsize + j
        ax.add_patch(plt.Rectangle(
            (j-0.5, i-0.5),
            1,
            1,
            facecolor=node_colors[inode],
            edgecolor="none",
            alpha=0.4,
            zorder=0,
        ))



"""
Module for edge arrows in iplotx.
"""

from typing import Never, Optional

from math import atan2, cos, sin
import numpy as np
import matplotlib as mpl
from matplotlib.patches import PathPatch

from ..style import (
    get_style,
    rotate_style,
)


class EdgeArrowCollection(mpl.collections.PatchCollection):
    """Collection of arrow patches for plotting directed edgs."""

    _factor = 1.0

    def __init__(
        self,
        edge_collection,
        *args,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        **kwargs,
    ) -> None:
        """Initialize the edge arrow collection.

        Parameters:
            edge_collection: The edge collection to which these arrows belong.
            transform: The transform to apply to the arrows. This related to the arrow size
                scaling, not the arrow tip position which is controlled by set_offset_transform.
        """

        self._edge_collection = edge_collection
        self._style = get_style(".edge.arrow")

        patches, sizes = self._create_artists()

        if "cmap" in self._edge_collection._style:
            kwargs["cmap"] = self._edge_collection._style["cmap"]
            kwargs["norm"] = self._edge_collection._style["norm"]

        super().__init__(
            patches,
            offsets=np.zeros((len(patches), 2)),
            offset_transform=self.get_offset_transform(),
            transform=transform,
            match_original=True,
            *args,
            **kwargs,
        )
        self._angles = np.zeros(len(self._paths))

        # Compute _transforms like in _CollectionWithScales for dpi issues
        self.set_sizes(sizes)

    def get_sizes(self):
        """Get arrow sizes (max of width and height), not scaled by dpi."""
        return self._sizes

    def get_sizes_dpi(self):
        """Get arrow sizes (max of width and height) scaled by dpi."""
        return self._transforms[:, 0, 0]

    def set_sizes(self, sizes, dpi=72.0):
        """Set vertex sizes.

        This rescales the current vertex symbol/path linearly, using this
        value as the largest of width and height.

        @param sizes: A sequence of vertex sizes or a single size.
        """
        if sizes is None:
            self._sizes = np.array([])
            self._transforms = np.empty((0, 3, 3))
        else:
            self._sizes = np.asarray(sizes)
            self._transforms = np.zeros((len(self._sizes), 3, 3))
            scale = self._sizes * dpi / 72.0 * self._factor
            self._transforms[:, 0, 0] = scale
            self._transforms[:, 1, 1] = scale
            self._transforms[:, 2, 2] = 1.0
        self.stale = True

    get_size = get_sizes
    set_size = set_sizes

    def set_figure(self, fig) -> None:
        """Set the figure for this artist and all children."""
        super().set_figure(fig)
        self._update_before_draw()

    def get_offset_transform(self):
        """Get offset transform for the edge arrows. This sets the tip of each arrow."""
        return self._edge_collection.get_transform()

    get_size = get_sizes
    set_size = set_sizes

    def _create_artists(self):
        style = self._style if self._style is not None else {}

        patches = []
        sizes = []
        for i in range(len(self._edge_collection._vertex_ids)):
            stylei = rotate_style(style, index=i)
            if ("facecolor" not in stylei) and ("color" not in stylei):
                stylei["facecolor"] = self._edge_collection.get_edgecolors()[i][:3]
            if ("edgecolor" not in stylei) and ("color" not in stylei):
                stylei["edgecolor"] = self._edge_collection.get_edgecolors()[i][:3]
            if "alpha" not in stylei:
                stylei["alpha"] = self._edge_collection.get_edgecolors()[i][3]
            if "linewidth" not in stylei:
                stylei["linewidth"] = self._edge_collection.get_linewidths()[i]

            patch, size = make_arrow_patch(
                **stylei,
            )
            patches.append(patch)
            sizes.append(size)

        return patches, sizes

    def _update_before_draw(self) -> None:
        """Update the arrow paths and directions before drawing, based on the edge collection."""
        self.set_sizes(self._sizes, self.get_figure(root=True).dpi)

        trans = self.get_offset_transform().transform

        for i, epath in enumerate(self._edge_collection.get_paths()):
            # Offset the arrow to point to the end of the edge
            self._offsets[i] = epath.vertices[-1]

            # Rotate the arrow to point in the direction of the edge
            apath = self._paths[i]
            # NOTE: because the tip of the arrow is at (0, 0) in patch space,
            # in theory it will rotate around that point already
            v2 = trans(epath.vertices[-1])
            v1 = trans(epath.vertices[-2])
            dv = v2 - v1
            theta = atan2(*(dv[::-1]))
            theta_old = self._angles[i]
            dtheta = theta - theta_old
            mrot = np.array([[cos(dtheta), sin(dtheta)], [-sin(dtheta), cos(dtheta)]])
            apath.vertices = apath.vertices @ mrot
            self._angles[i] = theta

    def set_array(self, A: np.ndarray) -> Never:
        """Set the array for cmap/norm coloring, but keep the facecolors as set (usually 'none')."""
        raise ValueError("Setting an array for arrows directly is not supported.")

    def set_colors(self, colors: np.ndarray) -> None:
        """Set arrow colors (edge and/or face) based on a colormap.

        Parameters:
            colors: Color array to apply. This must be an Nx3 or Nx4 vector of RGB or RGBA colors.
                This function will NOT attempt to convert other color descriptions to RGB/RGBA.
        """
        # NOTE: facecolors is always an array because we come from patches
        # It can have zero alpha (i.e. if we choose "none", or a hollow marker)
        self.set_edgecolor(colors)
        has_facecolor = self._facecolors[:, 3] > 0
        self._facecolors[has_facecolor] = colors[has_facecolor]

    @mpl.artist.allow_rasterization
    def draw(self, renderer):
        self._update_before_draw()
        super().draw(renderer)


def make_arrow_patch(
    marker: str = "|>",
    width: float = 8,
    height: Optional[float | str] = None,
    **kwargs,
):
    """Make a patch of the given marker shape and size.

    Parameters:
        marker: The marker shape to use. Supported markers are:
            "|>", "|/", "|\\", ">", "<", ">>", ")>", ")", "(", "]", "[", "|",
            "x", "s", "d", "p", "q". Dashes at the start of this string will
            be ignored, so "->" is equivalent to ">".
        width: The width of the marker in points. Height is by default 1.3 the
            width, unless specified separately.
        height: The height of the marker in points. If not specified, it is
            1.3 times the width. This can also be the string "width", in which
            case the height will be equal to the width.
        **kwargs: Additional keyword arguments passed to the PathPatch.

    Returns:
        A pair with the patch and the max size of the patch in points.
    """
    # Forget any leading dashes
    marker = marker.lstrip("-")

    if height is None:
        height = width * 1.3
    elif height == "width":
        height = width

    # Normalise by the max size, this is taken care of in _transforms
    # subsequently in a way that is nice to dpi scaling
    size_max = max(width, height)
    if size_max > 0:
        height /= size_max
        width /= size_max

    if marker == "|>":
        codes = ["MOVETO", "LINETO", "LINETO", "CLOSEPOLY"]
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        path = mpl.path.Path(
            np.array(
                [
                    [-height, width * 0.5],
                    [-height, -width * 0.5],
                    [0, 0],
                    [-height, width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "|\\":
        codes = ["MOVETO", "LINETO", "LINETO", "CLOSEPOLY"]
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        path = mpl.path.Path(
            np.array(
                [
                    [-height, width * 0.5],
                    [-height, 0],
                    [0, 0],
                    [-height, width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "|/":
        codes = ["MOVETO", "LINETO", "LINETO", "CLOSEPOLY"]
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        path = mpl.path.Path(
            np.array(
                [
                    [-height, 0],
                    [-height, -width * 0.5],
                    [0, 0],
                    [-height, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == ">":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO"]
        path = mpl.path.Path(
            np.array([[-height, width * 0.5], [0, 0], [-height, -width * 0.5]]),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "<":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO"]
        path = mpl.path.Path(
            np.array([[height, width * 0.5], [0, 0], [height, -width * 0.5]]),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == ">>":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        overhang = kwargs.pop("overhang", 0.25)
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [0, 0],
                    [-height, -width * 0.5],
                    [-height * (1.0 - overhang), 0],
                    [-height, width * 0.5],
                    [0, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == ")>":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        overhang = kwargs.pop("overhang", 0.25)
        codes = ["MOVETO", "LINETO", "CURVE3", "CURVE3", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [0, 0],
                    [-height, -width * 0.5],
                    [-height * (1.0 - overhang), 0],
                    [-height, width * 0.5],
                    [0, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == ")":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "CURVE3", "CURVE3"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height * 0.5, width * 0.5],
                    [height * 0.5, 0],
                    [-height * 0.5, -width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "(":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "CURVE3", "CURVE3"]
        path = mpl.path.Path(
            np.array(
                [
                    [height * 0.5, width * 0.5],
                    [-height * 0.5, 0],
                    [height * 0.5, -width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "]":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height, width * 0.5],
                    [0, width * 0.5],
                    [0, -width * 0.5],
                    [-height, -width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "[":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO"]
        path = mpl.path.Path(
            np.array(
                [
                    [height, width * 0.5],
                    [0, width * 0.5],
                    [0, -width * 0.5],
                    [height, -width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "|":
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO"]
        path = mpl.path.Path(
            np.array([[0, width * 0.5], [0, -width * 0.5]]),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker in ("x", "X"):
        kwargs["facecolor"] = "none"
        if "color" in kwargs:
            kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "MOVETO", "LINETO"]
        path = mpl.path.Path(
            np.array(
                [
                    [height * 0.5, width * 0.5],
                    [-height * 0.5, -width * 0.5],
                    [-height * 0.5, width * 0.5],
                    [height * 0.5, -width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=False,
        )
    elif marker == "s":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height, width * 0.5],
                    [-height, -width * 0.5],
                    [0, -width * 0.5],
                    [0, width * 0.5],
                    [-height, width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "d":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height * 0.5, width * 0.5],
                    [-height, 0],
                    [-height * 0.5, -width * 0.5],
                    [0, 0],
                    [-height * 0.5, width * 0.5],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "p":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height, 0],
                    [0, 0],
                    [0, -width],
                    [-height, -width],
                    [-height, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "q":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO", "LINETO", "LINETO", "LINETO", "CLOSEPOLY"]
        path = mpl.path.Path(
            np.array(
                [
                    [-height, 0],
                    [0, 0],
                    [0, width],
                    [-height, width],
                    [-height, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    elif marker == "none":
        if "color" in kwargs:
            kwargs["facecolor"] = kwargs["edgecolor"] = kwargs.pop("color")
        codes = ["MOVETO"]
        path = mpl.path.Path(
            np.array(
                [
                    [0, 0],
                ]
            ),
            codes=[getattr(mpl.path.Path, x) for x in codes],
            closed=True,
        )
    else:
        raise ValueError(f"Arrow marker not found: {marker}.")

    patch = PathPatch(
        path,
        **kwargs,
    )

    return patch, size_max



"""
This module focuses on how to ingest network/tree data into standard data structures no matter what library they come from.
"""

import pathlib
import pkgutil
import importlib
import warnings
from typing import (
    Optional,
    Sequence,
    Protocol,
)
from collections.abc import Hashable
import pandas as pd

from ..typing import (
    GraphType,
    LayoutType,
    TreeType,
)
from .typing import (
    NetworkDataProvider,
    NetworkData,
    TreeDataProvider,
    TreeData,
)

provider_protocols = {
    "network": NetworkDataProvider,
    "tree": TreeDataProvider,
}

# Internally supported data providers
data_providers: dict[str, dict[str, Protocol]] = {kind: {} for kind in provider_protocols}
for kind in data_providers:
    providers_path = pathlib.Path(__file__).parent.joinpath("providers").joinpath(kind)
    for importer, module_name, _ in pkgutil.iter_modules([providers_path]):
        module = importlib.import_module(f"iplotx.ingest.providers.{kind}.{module_name}")
        for key, val in module.__dict__.items():
            if key == provider_protocols[kind].__name__:
                continue
            if key.endswith("DataProvider"):
                data_providers[kind][module_name] = val
                break
    del providers_path


def network_library(network) -> str:
    """Guess the network library used to create the network."""
    for name, provider in data_providers["network"].items():
        if provider.check_dependencies():
            graph_type = provider.graph_type()
            if isinstance(network, graph_type):
                return name
    raise ValueError(
        f"Network {network} did not match any available network library.",
    )


def tree_library(tree) -> str:
    """Guess the tree library used to create the tree."""
    for name, provider in data_providers["tree"].items():
        if provider.check_dependencies():
            tree_type = provider.tree_type()
            if isinstance(tree, tree_type):
                return name
    raise ValueError(
        f"Tree {tree} did not match any available tree library.",
    )


# Functions to ingest data from various libraries
def ingest_network_data(
    network: GraphType,
    layout: Optional[LayoutType] = None,
    vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
    edge_labels: Optional[Sequence[str] | dict[str,]] = None,
) -> NetworkData:
    """Create internal data for the network."""
    _update_data_providers("network")

    nl = network_library(network)

    if nl in data_providers["network"]:
        provider: NetworkDataProvider = data_providers["network"][nl]
    else:
        sup = ", ".join(data_providers["network"].keys())
        raise ValueError(
            f"Network library '{nl}' is not installed. "
            f"Currently installed supported libraries: {sup}."
        )

    result = provider(network)(
        layout=layout,
        vertex_labels=vertex_labels,
        edge_labels=edge_labels,
    )
    result["network_library"] = nl
    return result


def ingest_tree_data(
    tree: TreeType,
    layout: Optional[str] = "horizontal",
    directed: bool | str = False,
    layout_style: Optional[dict[str, str | int | float]] = None,
    vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
    edge_labels: Optional[Sequence[str] | dict[Hashable, str]] = None,
    leaf_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
) -> TreeData:
    """Create internal data for the tree."""
    _update_data_providers("tree")

    tl = tree_library(tree)

    if tl in data_providers["tree"]:
        provider: TreeDataProvider = data_providers["tree"][tl]
    else:
        sup = ", ".join(data_providers["tree"].keys())
        raise ValueError(
            f"Tree library '{tl}' is not installed. Currently installed supported libraries: {sup}."
        )

    result = provider(
        tree=tree,
    )(
        layout=layout,
        directed=directed,
        layout_style=layout_style,
        vertex_labels=vertex_labels,
        edge_labels=edge_labels,
        leaf_labels=leaf_labels,
    )
    result["tree_library"] = tl

    return result


# INTERNAL FUNCTIONS
def _update_data_providers(kind):
    """Update data provieders dynamically from external packages."""
    discovered_providers = importlib.metadata.entry_points(group=f"iplotx.{kind}_data_providers")
    for entry_point in discovered_providers:
        if entry_point.name not in data_providers["network"]:
            try:
                data_providers[kind][entry_point.name] = entry_point.load()
            except Exception as e:
                warnings.warn(f"Failed to load {kind} data provider '{entry_point.name}': {e}")
